@@ -30,21 +30,17 @@ import weakref
 
 from requests.exceptions import RequestException, ProxyError
 from fsspec.asyn import sync_wrapper, sync, AsyncFileSystem
-from fsspec.utils import stringify_path
+from fsspec.utils import stringify_path, setup_logging
 from fsspec.implementations.http import get_client
 from .utils import ChecksumError, HttpError, is_retriable
 from .checkers import get_consistency_checker, MD5Checker
 from . import __version__ as version
 
 logger = logging.getLogger("gcsfs")
+
+
 if "GCSFS_DEBUG" in os.environ:
-    handle = logging.StreamHandler()
-    formatter = logging.Formatter(
-        "%(asctime)s - %(name)s - %(levelname)s " "- %(message)s"
-    )
-    handle.setFormatter(formatter)
-    logger.addHandler(handle)
-    logger.setLevel("DEBUG")
+    setup_logging(logger=logger, level=os.environ["GCSFS_DEBUG"])
 
 
 # client created 2018-01-16
@@ -80,6 +76,7 @@ bACLs = {
 DEFAULT_PROJECT = os.environ.get("GCSFS_DEFAULT_PROJECT", "")
 
 GCS_MIN_BLOCK_SIZE = 2 ** 18
+GCS_MAX_BLOCK_SIZE = 2 ** 28
 DEFAULT_BLOCK_SIZE = 5 * 2 ** 20
 
 
@@ -117,6 +114,15 @@ async def _req_to_text(r):
         return (await r.read()).decode()
 
 
+def unjson(contents):
+    import ujson
+
+    try:
+        return ujson.loads(contents)
+    except:  # noqa: E722
+        return None
+
+
 class GCSFileSystem(AsyncFileSystem):
     r"""
     Connect to Google Cloud Storage.
@@ -236,7 +242,7 @@ class GCSFileSystem(AsyncFileSystem):
         requester_pays=False,
         asynchronous=False,
         loop=None,
-        callback_timeout=None,
+        timeout=None,
         **kwargs,
     ):
         super().__init__(
@@ -264,16 +270,32 @@ class GCSFileSystem(AsyncFileSystem):
         self.cache_timeout = cache_timeout or kwargs.pop("listings_expiry_time", None)
         self.requests_timeout = requests_timeout
         self.check_credentials = check_connection
-        self.callback_timeout = callback_timeout
-        if not asynchronous:
-            self._session = sync(
-                self.loop, get_client, callback_timeout=self.callback_timeout
-            )
-            weakref.finalize(self, sync, self.loop, self.session.close)
-        else:
-            self._session = None
+        self.timeout = timeout
+        self._session = None
+        if not self.asynchronous:
+            self._session = sync(self.loop, get_client, timeout=self.timeout)
+            weakref.finalize(self, self.close_session, self.loop, self._session)
         self.connect(method=token)
 
+    @staticmethod
+    def close_session(loop, session):
+        if loop is not None and session is not None:
+            if loop.is_running():
+                sync(loop, session.close, timeout=0.1)
+            else:
+                pass
+
+    async def _set_session(self):
+        if self._session is None:
+            self._session = await get_client()
+        return self._session
+
+    @property
+    def session(self):
+        if self.asynchronous and self._session is None:
+            raise RuntimeError("Please await _connect* before anything else")
+        return self._session
+
     @classmethod
     def _strip_protocol(cls, path):
         if isinstance(path, list):
@@ -288,17 +310,6 @@ class GCSFileSystem(AsyncFileSystem):
         # use of root_marker to make minimum required path, e.g., "/"
         return path or cls.root_marker
 
-    @property
-    def session(self):
-        if self._session is None:
-            raise RuntimeError("please await ``.set_session`` before anything else")
-        return self._session
-
-    async def set_session(self):
-        from fsspec.implementations.http import get_client
-
-        self._session = await get_client()
-
     @classmethod
     def load_tokens(cls):
         """Get "browser" tokens from disc"""
@@ -400,6 +411,7 @@ class GCSFileSystem(AsyncFileSystem):
         with self.lock:
             if self.credentials.valid:
                 return  # repeat to avoid race (but don't want lock in common case)
+            logger.debug("GCS refresh")
             self.credentials.refresh(req)
             self.credentials.apply(self.heads)
 
@@ -498,7 +510,7 @@ class GCSFileSystem(AsyncFileSystem):
     async def _call(
         self, method, path, *args, json_out=False, info_out=False, **kwargs
     ):
-        logger.debug(f"{method.upper()}: {path}, {args}, {kwargs}")
+        logger.debug(f"{method.upper()}: {path}, {args}, {kwargs.get('headers')}")
 
         for retry in range(self.retries):
             try:
@@ -508,6 +520,7 @@ class GCSFileSystem(AsyncFileSystem):
                 path, jsonin, datain, headers, kwargs = self._get_args(
                     path, *args, **kwargs
                 )
+                await self._set_session()
                 async with self.session.request(
                     method=method,
                     url=path,
@@ -517,21 +530,15 @@ class GCSFileSystem(AsyncFileSystem):
                     data=datain,
                     timeout=self.requests_timeout,
                 ) as r:
-                    import json
 
                     status = r.status
                     headers = r.headers
                     info = r.request_info  # for debug only
                     contents = await r.read()
-                    try:
-                        json = json.loads(contents)
-                    except Exception:
-                        json = None
 
-                self.validate_response(status, contents, json, path, headers)
+                self.validate_response(status, contents, path, headers)
                 break
             except (HttpError, RequestException, GoogleAuthError, ChecksumError) as e:
-                json = None
                 if (
                     isinstance(e, HttpError)
                     and e.code == 400
@@ -548,7 +555,7 @@ class GCSFileSystem(AsyncFileSystem):
                 logger.exception("_call non-retriable exception: %s" % e)
                 raise e
         if json_out:
-            return json
+            return unjson(contents)
         elif info_out:
             return info
         else:
@@ -560,10 +567,7 @@ class GCSFileSystem(AsyncFileSystem):
     def buckets(self):
         """Return list of available project buckets."""
         return [
-            b["name"]
-            for b in sync(
-                self.loop, self._list_buckets(), callback_timeout=self.callback_timeout
-            )
+            b["name"] for b in sync(self.loop, self._list_buckets, timeout=self.timeout)
         ]
 
     @staticmethod
@@ -844,7 +848,7 @@ class GCSFileSystem(AsyncFileSystem):
         else:
             raise FileNotFoundError(path)
 
-    def glob(self, path, prefix="", **kwargs):
+    async def _glob(self, path, prefix="", **kwargs):
         if not prefix:
             # Identify pattern prefixes. Ripped from fsspec.spec.AbstractFileSystem.glob and matches
             # the glob.has_magic patterns.
@@ -854,7 +858,7 @@ class GCSFileSystem(AsyncFileSystem):
 
             ind = min(indstar, indques, indbrace)
             prefix = path[:ind].split("/")[-1]
-        return super().glob(path, prefix=prefix, **kwargs)
+        return await super()._glob(path, prefix=prefix, **kwargs)
 
     async def _ls(self, path, detail=False, prefix="", **kwargs):
         """List objects under the given '/{bucket}/{prefix} path."""
@@ -888,11 +892,13 @@ class GCSFileSystem(AsyncFileSystem):
         headers, out = await self._call("GET", u2, headers=head)
         return out
 
-    def getxattr(self, path, attr):
+    async def _getxattr(self, path, attr):
         """Get user-defined metadata attribute"""
-        meta = self.info(path).get("metadata", {})
+        meta = (await self._info(path)).get("metadata", {})
         return meta[attr]
 
+    getxattr = sync_wrapper(_getxattr)
+
     async def _setxattrs(
         self, path, content_type=None, content_encoding=None, **kwargs
     ):
@@ -982,7 +988,7 @@ class GCSFileSystem(AsyncFileSystem):
                 json_out=True,
             )
 
-    async def _rm_one_file(self, path):
+    async def _rm_file(self, path):
         bucket, key = self.split_path(path)
         if key:
             await self._call("DELETE", "b/{}/o/{}", bucket, key)
@@ -1033,7 +1039,8 @@ class GCSFileSystem(AsyncFileSystem):
             out = set(re.findall(pattern, txt))
             raise OSError(out)
 
-    async def _rm(self, paths, batchsize):
+    async def _rm(self, path, recursive=False, maxdepth=None, batchsize=20):
+        paths = await self._expand_path(path, recursive=recursive, maxdepth=maxdepth)
         files = [p for p in paths if self.split_path(p)[1]]
         dirs = [p for p in paths if not self.split_path(p)[1]]
         exs = await asyncio.gather(
@@ -1050,6 +1057,8 @@ class GCSFileSystem(AsyncFileSystem):
             raise exs[0]
         await asyncio.gather(*[self._rmdir(d) for d in dirs])
 
+    rm = sync_wrapper(_rm)
+
     async def _pipe_file(
         self,
         path,
@@ -1134,25 +1143,19 @@ class GCSFileSystem(AsyncFileSystem):
         except IOError:
             return False
 
-    def find(self, path, withdirs=False, detail=False, prefix="", **kwargs):
+    async def _find(self, path, withdirs=False, detail=False, prefix="", **kwargs):
         path = self._strip_protocol(path)
         bucket, key = self.split_path(path)
-        out, _ = sync(
-            self.loop,
-            self._do_list_objects,
+        out, _ = await self._do_list_objects(
             path,
             delimiter=None,
             prefix=prefix,
-            callback_timeout=self.callback_timeout,
         )
         if not out and key:
             try:
                 out = [
-                    sync(
-                        self.loop,
-                        self._get_object,
+                    await self._get_object(
                         path,
-                        callback_timeout=self.callback_timeout,
                     )
                 ]
             except FileNotFoundError:
@@ -1161,24 +1164,27 @@ class GCSFileSystem(AsyncFileSystem):
         sdirs = set()
         cache_entries = {}
         for o in out:
-            par = self._parent(o["name"])
-            if par not in sdirs:
-                sdirs.add(par)
-                dirs.append(
-                    {
-                        "Key": self.split_path(par)[1],
-                        "Size": 0,
-                        "name": par,
-                        "StorageClass": "DIRECTORY",
-                        "type": "directory",
-                        "size": 0,
-                    }
-                )
-            # Don't cache "folder-like" objects (ex: "Create Folder" in GCS console) to prevent
-            # masking subfiles in subsequent requests.
-            if not o["name"].endswith("/"):
-                cache_entries.setdefault(par, [])
-                cache_entries[par].append(o)
+            par = o["name"]
+            while par:
+                par = self._parent(par)
+                if par not in sdirs:
+                    if len(par) < len(path):
+                        break
+                    sdirs.add(par)
+                    dirs.append(
+                        {
+                            "Key": self.split_path(par)[1],
+                            "Size": 0,
+                            "name": par,
+                            "StorageClass": "DIRECTORY",
+                            "type": "directory",
+                            "size": 0,
+                        }
+                    )
+                # Don't cache "folder-like" objects (ex: "Create Folder" in GCS console) to prevent
+                # masking subfiles in subsequent requests.
+                if not o["name"].endswith("/"):
+                    cache_entries.setdefault(par, []).append(o)
         self.dircache.update(cache_entries)
 
         if withdirs:
@@ -1218,7 +1224,7 @@ class GCSFileSystem(AsyncFileSystem):
             os.makedirs(os.path.dirname(lpath), exist_ok=True)
             with open(lpath, "wb") as f2:
                 while True:
-                    data = await r.content.read(4096)
+                    data = await r.content.read(4096 * 32)
                     if not data:
                         break
                     f2.write(data)
@@ -1226,16 +1232,6 @@ class GCSFileSystem(AsyncFileSystem):
 
             checker.validate_http_response(r)
 
-    def rm(self, path, recursive=False, batchsize=20):
-        paths = self.expand_path(path, recursive=recursive)
-        sync(
-            self.loop,
-            self._rm,
-            paths,
-            callback_timeout=self.callback_timeout,
-            batchsize=batchsize,
-        )
-
     def _open(
         self,
         path,
@@ -1291,7 +1287,7 @@ class GCSFileSystem(AsyncFileSystem):
         else:
             return path.split("/", 1)
 
-    def validate_response(self, status, content, json, path, headers=None):
+    def validate_response(self, status, content, path, headers=None):
         """
         Check the requests object r, raise error if it's not ok.
 
@@ -1302,8 +1298,10 @@ class GCSFileSystem(AsyncFileSystem):
         """
         if status >= 400:
             error = None
+            if hasattr(content, "decode"):
+                content = content.decode()
             try:
-                error = json["error"]
+                error = unjson(content)["error"]
                 msg = error["message"]
             except:  # noqa: E722
                 # TODO: limit to appropriate exceptions
@@ -1320,11 +1318,18 @@ class GCSFileSystem(AsyncFileSystem):
             elif error:
                 raise HttpError(error)
             elif status:
-                raise HttpError({"code": status})
+                raise HttpError(
+                    {
+                        "code": status,
+                        "message": msg,
+                    }
+                )  # text-like
             else:
                 raise RuntimeError(msg)
         else:
-            if headers is not None and "X-Goog-Hash" in headers:
+            if self.consistency != "md5":
+                return None
+            elif headers is not None and "X-Goog-Hash" in headers:
                 checker = MD5Checker()
                 checker.update(content)
                 checker.validate_headers(headers)
@@ -1347,7 +1352,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         consistency="md5",
         metadata=None,
         content_type=None,
-        callback_timeout=None,
+        timeout=None,
         **kwargs,
     ):
         """
@@ -1376,7 +1381,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             content types at https://www.iana.org/assignments/media-types/media-types.txt
         metadata: dict
             Custom metadata, in key/value pairs, added at file creation
-        callback_timeout: int
+        timeout: int
             Timeout seconds for the asynchronous callback.
         """
         super().__init__(
@@ -1395,11 +1400,15 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         self.gcsfs = gcsfs
         self.bucket = bucket
         self.key = key
-        self.metadata = metadata
         self.acl = acl
         self.checker = get_consistency_checker(consistency)
-        self.content_type = content_type or "application/octet-stream"
-        self.callback_timeout = callback_timeout
+
+        det = getattr(self, "details", {})  # only exists in read mode
+        self.content_type = content_type or det.get(
+            "contentType", "application/octet-stream"
+        )
+        self.metadata = metadata or det.get("metadata", {})
+        self.timeout = timeout
         if mode == "wb":
             if self.blocksize < GCS_MIN_BLOCK_SIZE:
                 warnings.warn("Setting block size to minimum value, 2**18")
@@ -1422,51 +1431,66 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         final: bool
             Complete and commit upload
         """
-        data = self.buffer.getvalue()
-        head = {}
-        l = len(data)
-        if final and self.autocommit:
-            if l:
-                # last chunk
-                head["Content-Range"] = "bytes %i-%i/%i" % (
-                    self.offset,
-                    self.offset + l - 1,
-                    self.offset + l,
-                )
-            else:
-                # closing when buffer is empty
-                head["Content-Range"] = "bytes */%i" % self.offset
-                data = None
-        else:
-            if l < GCS_MIN_BLOCK_SIZE:
-                if not self.autocommit:
-                    return False
-                elif not final:
-                    raise ValueError("Non-final chunk write below min size.")
-            head["Content-Range"] = "bytes %i-%i/*" % (self.offset, self.offset + l - 1)
-        head.update({"Content-Type": self.content_type, "Content-Length": str(l)})
-        headers, contents = self.gcsfs.call(
-            "POST", self.location, headers=head, data=data
-        )
-        if "Range" in headers:
-            end = int(headers["Range"].split("-")[1])
-            shortfall = (self.offset + l - 1) - end
-            if shortfall:
-                self.checker.update(data[:-shortfall])
-                self.buffer = io.BytesIO(data[-shortfall:])
-                self.buffer.seek(shortfall)
-                self.offset += l - shortfall
+        while True:
+            # shortfall splits blocks bigger than max allowed upload
+            data = self.buffer.getvalue()
+            head = {}
+            l = len(data)
+
+            if (l < GCS_MIN_BLOCK_SIZE) and not final:
+                # either flush() was called, but we don't have enough to
+                # push, or we split a big upload, and have less left than one
+                # block.  If this is the final part, OK to violate those
+                # terms.
                 return False
+
+            # Select the biggest possible chunk of data to be uploaded
+            chunk_length = min(l, GCS_MAX_BLOCK_SIZE)
+            chunk = data[:chunk_length]
+            if final and self.autocommit and chunk_length == l:
+                if l:
+                    # last chunk
+                    head["Content-Range"] = "bytes %i-%i/%i" % (
+                        self.offset,
+                        self.offset + chunk_length - 1,
+                        self.offset + l,
+                    )
+                else:
+                    # closing when buffer is empty
+                    head["Content-Range"] = "bytes */%i" % self.offset
+                    data = None
             else:
-                self.checker.update(data)
-        elif l:
-            #
-            assert final, "Response looks like upload is over"
-            j = json.loads(contents)
-            self.checker.update(data)
-            self.checker.validate_json_response(j)
-        else:
-            assert final, "Response looks like upload is over"
+                head["Content-Range"] = "bytes %i-%i/*" % (
+                    self.offset,
+                    self.offset + chunk_length - 1,
+                )
+            head.update(
+                {"Content-Type": self.content_type, "Content-Length": str(chunk_length)}
+            )
+            headers, contents = self.gcsfs.call(
+                "POST", self.location, headers=head, data=chunk
+            )
+            if "Range" in headers:
+                end = int(headers["Range"].split("-")[1])
+                shortfall = (self.offset + l - 1) - end
+                if shortfall:
+                    self.checker.update(data[:-shortfall])
+                    self.buffer = io.BytesIO(data[-shortfall:])
+                    self.buffer.seek(shortfall)
+                    self.offset += l - shortfall
+                    continue
+                else:
+                    self.checker.update(data)
+            else:
+                assert final, "Response looks like upload is over"
+                if l:
+                    j = json.loads(contents)
+                    self.checker.update(data)
+                    self.checker.validate_json_response(j)
+            # Clear buffer and update offset when all is received
+            self.buffer = io.BytesIO()
+            self.offset += l
+            break
         return True
 
     def commit(self):
@@ -1484,7 +1508,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             self.key,
             self.content_type,
             self.metadata,
-            callback_timeout=self.callback_timeout,
+            timeout=self.timeout,
         )
 
     def discard(self):
@@ -1517,7 +1541,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             self.metadata,
             self.consistency,
             self.content_type,
-            callback_timeout=self.callback_timeout,
+            timeout=self.timeout,
         )
 
     def _fetch_range(self, start=None, end=None):
@@ -1552,7 +1576,9 @@ async def upload_chunk(fs, location, data, offset, size, content_type):
         end = int(headers["Range"].split("-")[1])
         shortfall = (offset + l - 1) - end
         if shortfall:
-            return await upload_chunk(fs, data[-shortfall:], end, size, content_type)
+            return await upload_chunk(
+                fs, location, data[-shortfall:], end, size, content_type
+            )
     return json.loads(txt) if txt else None
 
 
