@@ -48,11 +48,12 @@ def test_simple():
     gcs.ls("/" + TEST_BUCKET)  # OK to lead with '/'
 
 
-@pytest.mark.skipif(ON_VCR, reason="async fail")
 @my_vcr.use_cassette(match=["all"])
 def test_many_connect():
     from multiprocessing.pool import ThreadPool
 
+    GCSFileSystem(TEST_PROJECT, token=GOOGLE_TOKEN)
+
     def task(i):
         GCSFileSystem(TEST_PROJECT, token=GOOGLE_TOKEN).ls("")
         return True
@@ -64,6 +65,22 @@ def test_many_connect():
     pool.join()
 
 
+@my_vcr.use_cassette(match=["all"])
+def test_many_connect_new():
+    from multiprocessing.pool import ThreadPool
+
+    def task(i):
+        # first instance is made within thread - creating loop
+        GCSFileSystem(TEST_PROJECT, token=GOOGLE_TOKEN).ls("")
+        return True
+
+    pool = ThreadPool(processes=20)
+    out = pool.map(task, range(40))
+    assert all(out)
+    pool.close()
+    pool.join()
+
+
 @my_vcr.use_cassette(match=["all"])
 def test_simple_upload():
     with gcs_maker() as gcs:
@@ -75,6 +92,21 @@ def test_simple_upload():
         assert gcs.cat(fn) == b"zz"
 
 
+@my_vcr.use_cassette(match=["all"])
+def test_large_upload():
+    orig = gcsfs.core.GCS_MAX_BLOCK_SIZE
+    gcsfs.core.GCS_MAX_BLOCK_SIZE = 262144  # minimum block size
+    try:
+        with gcs_maker() as gcs:
+            fn = TEST_BUCKET + "/test"
+            d = b"7123" * 262144
+            with gcs.open(fn, "wb", content_type="application/octet-stream") as f:
+                f.write(d)
+            assert gcs.cat(fn) == d
+    finally:
+        gcsfs.core.GCS_MAX_BLOCK_SIZE = orig
+
+
 @pytest.mark.xfail(reason="oddness in repeat VCR calls")
 @my_vcr.use_cassette(match=["all"])
 def test_multi_upload():
@@ -952,26 +984,26 @@ def test_raise_on_project_mismatch(mock_auth):
 
 
 def test_validate_response():
-    gcs = GCSFileSystem(token="anon")
-    gcs.validate_response(200, None, None, "/path")
+    gcs = GCSFileSystem(token="anon", consistency=None)
+    gcs.validate_response(200, None, "/path")
 
     # HttpError with no JSON body
     with pytest.raises(HttpError) as e:
-        gcs.validate_response(503, b"", None, "/path")
+        gcs.validate_response(503, b"", "/path")
     assert e.value.code == 503
     assert e.value.message == ", 503"
 
     # HttpError with JSON body
-    j = {"error": {"code": 503, "message": b"Service Unavailable"}}
+    j = '{"error": {"code": 503, "message": "Service Unavailable"}}'
     with pytest.raises(HttpError) as e:
-        gcs.validate_response(503, None, j, "/path")
+        gcs.validate_response(503, j, "/path")
     assert e.value.code == 503
-    assert e.value.message == b"Service Unavailable, 503"
+    assert e.value.message == "Service Unavailable, 503"
 
     # 403
-    j = {"error": {"message": "Not ok"}}
+    j = '{"error": {"message": "Not ok"}}'
     with pytest.raises(IOError, match="Forbidden: /path\nNot ok"):
-        gcs.validate_response(403, None, j, "/path")
+        gcs.validate_response(403, j, "/path")
 
     # 404
     with pytest.raises(FileNotFoundError):
@@ -981,10 +1013,13 @@ def test_validate_response():
     with pytest.raises(ProxyError):
         gcs.validate_response(502, b"", None, "/path")
 
-    # ChecksumError
+    # No response validation when the method isn't md5
     md5 = repr(base64.b64encode(hashlib.md5(b"foo").digest()))[2:-1]
+    gcs.validate_response(0, b"f", "/path", {"X-Goog-Hash": f"md5={md5}"})
+
+    gcs.consistency = "md5"
     with pytest.raises(ChecksumError):
-        gcs.validate_response(0, b"f", None, "/path", {"X-Goog-Hash": f"md5={md5}"})
+        gcs.validate_response(0, b"f", "/path", {"X-Goog-Hash": f"md5={md5}"})
 
 
 @my_vcr.use_cassette(match=["all"])
@@ -1029,12 +1064,12 @@ def test_metadata_read_permissions(
     with gcs_maker(True) as gcs:
         _validate_response = gcs.validate_response
 
-        def validate_response(self, status, content, json, path, headers=None):
+        def validate_response(self, status, content, path, headers=None):
             if path.endswith(f"/o{file_path}") and validate_get_error is not None:
                 raise validate_get_error
             if path.endswith("/o/") and validate_list_error is not None:
                 raise validate_list_error
-            _validate_response(status, content, json, path, headers=None)
+            _validate_response(status, content, path, headers=None)
 
         gcs.validate_response = validate_response.__get__(gcs)
 
@@ -1071,6 +1106,22 @@ def test_placeholder_dir_cache_validity():
         gcs.info(f"gs://{TEST_BUCKET}/b")
 
 
+@my_vcr.use_cassette(match=["all"])
+def test_pseudo_dir_find():
+    with gcs_maker(False) as fs:
+        fs.touch(f"{TEST_BUCKET}/a/b/file")
+        b = set(fs.glob(f"{TEST_BUCKET}/a/*"))
+        assert f"{TEST_BUCKET}/a/b" in b
+        a = set(fs.glob(f"{TEST_BUCKET}/*"))
+        assert f"{TEST_BUCKET}/a" in a
+        assert fs.find(TEST_BUCKET) == [f"{TEST_BUCKET}/a/b/file"]
+        assert fs.find(f"{TEST_BUCKET}/a", withdirs=True) == [
+            f"{TEST_BUCKET}/a",
+            f"{TEST_BUCKET}/a/b",
+            f"{TEST_BUCKET}/a/b/file",
+        ]
+
+
 @my_vcr.use_cassette(match=["all"])
 def test_zero_cache_timeout():
     with gcs_maker(True, cache_timeout=0) as gcs:
